@@ -491,18 +491,36 @@ class Testbed:
             adtlog.debug("Copying file %s to testbed %s" % (host, tb))
             TestbedPath(self, host, tb, os.path.isdir(host)).copydown()
 
-        # Get the testbed release.
-        self.release = self._get_release()
-        if not self.release and (self.add_apt_releases or self.add_apt_pockets):
-            self.bomb(
-                "Could not guess the testbed release, but adding releases or pockets require it",
-                adtlog.AutopkgtestError,
-            )
-
         # Pin the default release if explicitly requested.
         if self.default_release:
             self._set_default_release()
 
+        if (
+            self.apt_upgrade
+            or self.add_apt_sources
+            or self.add_apt_pockets
+            or self.add_apt_releases
+        ):
+            # Update APT package index. We can't run this unconditionally because
+            # we want to support not-root-on-testned and offline use cases.
+            adtlog.info("updating testbed package index (apt update)")
+            try:
+                self._run_apt_command(action="update")
+            except adtlog.AptError:
+                self.bomb("failed testbed apt update")
+
+        if self.add_apt_releases or self.add_apt_pockets:
+            # Assume the testbed release to be the one that provides the candidate
+            # base-files version. This requires the package index to be up-to-date.
+            _, _, self.release = self.get_candidate("base-files")
+            if self.release:
+                adtlog.info(f"testbed release detected to be: {self.release}")
+            else:
+                self.bomb(
+                    "could not determine the testbed release (no APT source for base-files?)",
+                    adtlog.AutopkgtestError,
+                )
+
         pin_packages = self.pin_packages.copy()
         add_apt_releases = self.add_apt_releases.copy()
         for pocket in self.add_apt_pockets:
@@ -1823,7 +1841,7 @@ class Testbed:
                 apt_cache = self.check_exec(["apt-cache", "dumpavail"], True)
                 for paragr in Deb822.iter_paragraphs(apt_cache):
                     pkg = paragr.get("Package")
-                    src = paragr.get("Source", pkg)
+                    src = paragr.get("Source", pkg).split()[0]
                     if src in srcpkgs and pkg not in binpkgs:
                         binpkgs.append(pkg)
 
@@ -1857,24 +1875,6 @@ class Testbed:
 
         self.apt_pin_for_releases.append(release)
 
-    def _get_release(self) -> Optional[str]:
-        """Get the release name which occurs in apt sources."""
-
-        script = load_shell_script("setup-commands/get-release")
-        rc, out, err = self.execute(
-            ["sh", "-ec", script],
-            stdout=subprocess.PIPE,
-            stderr=subprocess.PIPE,
-        )
-        if rc != 0:
-            err = err.strip()
-            adtlog.warning(f"Failed to get the testbed release: {err}")
-            return None
-
-        release = out.strip()
-        adtlog.info(f"testbed release detected to be: {self.release}")
-        return release
-
     def _set_default_release(self):
         """Set apt's default release."""
         # The "default release" preference file should sort first so to behave
@@ -1899,11 +1899,12 @@ class Testbed:
         if filename.endswith(".sources"):
             # add empty line to ensure deb822 stanzas are correctly separated
             source += "\n"
+        new_source_path = f"/etc/apt/sources.list.d/{filename}"
         self.check_exec(
             [
                 "sh",
                 "-ec",
-                f'"$@" >> "/etc/apt/sources.list.d/{filename}"',
+                f'"$@" >> "{new_source_path}"',
                 "add_apt_source",
                 "printf",
                 r"%s\n",
@@ -1911,6 +1912,20 @@ class Testbed:
             ]
         )
 
+        adtlog.info(f"updating testbed package index for new source {filename}")
+        try:
+            self._run_apt_command(
+                action="update",
+                # Options to only update the added source, without discarding existing data.
+                extra_opts=[
+                    f"-o=Dir::Etc::SourceList={new_source_path}",
+                    "-o=Dir::Etc::SourceParts=/dev/null",
+                    "-oAPT::Get::List-Cleanup=0",
+                ],
+            )
+        except adtlog.AptError:
+            self.bomb(f"failed testbed apt update for {filename}")
+
     def _get_apt_sources_one_line_style(self) -> List[str]:
         """Return list of all the configured one-line-style sources."""
 
@@ -2207,12 +2222,14 @@ class Testbed:
 
         return stdout
 
-    def get_candidate(self, package: str) -> Tuple[Optional[Version], Optional[int]]:
+    def get_candidate(
+        self, package: str
+    ) -> Tuple[Optional[Version], Optional[int], Optional[str]]:
         """Return candidate version and priority for package"""
 
         policy = self.get_policy_for_package(package)
         if not policy:
-            return None, None
+            return None, None, None
 
         # Get the candidate version of this binary
         version = None
@@ -2225,21 +2242,23 @@ class Testbed:
             break
 
         if not version:
-            return None, None
+            return None, None, None
 
-        priority = self._get_priority_from_policy(policy, version)
+        priority, suite = self._version_info_from_policy(policy, version)
 
-        return Version(version), priority
+        return Version(version), priority, suite
 
-    def _get_priority_from_policy(self, policy: str, version: str) -> int:
+    def _version_info_from_policy(
+        self, policy: str, version: str
+    ) -> Tuple[int, Optional[str]]:
         """Extract the priority for a package version from its policy.
 
         :param policy: `apt-cache policy` output for a single package
         :type policy: str
         :param version: the version for which to extract the prioriy
         :type version: str
-        :return: priority for the given version
-        :rtype: int
+        :return: priority and suite for the given version
+        :rtype: tuple[int, str]
         """
 
         # The apt-policy output for a package looks like this:
@@ -2295,22 +2314,37 @@ class Testbed:
             "version_table doesn't contain version '%s':\n%s\npolicy:\n%s"
             % (version, version_table, policy)
         )
-        version_table_for_candidate = needle.group(0)
+        version_table_for_candidate = needle.group(0).splitlines()
 
         # 3. Extract the "version line" and from that the priority.
-        version_line = version_table_for_candidate.splitlines()[0]
+        version_line = version_table_for_candidate[0]
         priority = int(version_line[5:].split()[1])
 
-        # This will always be true with apt >= 1.1~exp9.
-        if priority != 0:
-            return priority
+        # This will always be false with apt >= 1.1~exp9.
+        if priority == 0:
+            # 4. Fallback to parsing "file priorities".
+            file_priorities = version_table_for_candidate[1:]
+            priorities = [int(fp.split(maxsplit=1)[0]) for fp in file_priorities]
+            priority = max(priorities)
 
-        # 5. Fallback to parsing "file priorities".
-        file_priorities = version_table_for_candidate.splitlines()[1:]
-        priorities = [int(fp.split(maxsplit=1)[0]) for fp in file_priorities]
-        priority = max(priorities)
-        assert priority != 0, "invalid priority"
-        return priority
+        assert priority != 0, "invalid priority, bug in policy parsing?"
+
+        # Get suite from first highest-priority file entry providing the candidate.
+        suite = None
+        max_file_priority = float("-inf")
+        for line in version_table_for_candidate[1:]:
+            file_parts = line.split()
+            if len(file_parts) < 5:
+                # Skip /var/lib/dpkg/status and other possibly malformed entries.
+                continue
+            # Use a strict > and not a >= so that we respect source order.
+            this_file_priority = int(file_parts[0])
+            if this_file_priority > max_file_priority:
+                max_file_priority = this_file_priority
+                suite = file_parts[2].split("/")[0].split("-")[0]
+                assert suite, "invalid suite, bug in policy parsing?"
+
+        return priority, suite
 
 
 class TestbedPath:
